#include <libregexis024vm/libregexis024vm.h>
#include <libregexis024vm/instruction_implementation.h>
#include <utility>

regex_sslot_id_t REGEX_IS024_Stack::pop() {
    assert(sz != 0);
    return slots[--sz];
}

void REGEX_IS024_Stack::append(regex_sslot_id_t x) {
    assert(slots);
    slots[sz] = x;
    sz++;
}

bool REGEX_IS024_Stack::empty() const {
    return !non_empty();
}

bool REGEX_IS024_Stack::non_empty() const {
    return sz;
}

REGEX_IS024_Stack::~REGEX_IS024_Stack() {
    assert(empty());
    free(slots);
}

REGEX_IS024_CONTEXT::REGEX_IS024_CONTEXT(size_t programSize, const uint8_t *data,
                                         uint64_t caTreeLimit, regex_tai_t saLenLimit,
                                         regex_sslot_id_t readSsLimit, regex_sslot_id_t forkSsLimit,
                                         uint64_t timeTickLimit) :
        program_size(programSize), prg(data), CA_TREE_LIMIT(caTreeLimit), SA_LEN_LIMIT(saLenLimit),
        READ_SS_LIMIT(readSsLimit), FORK_SS_LIMIT(forkSsLimit), time_tick_limit(timeTickLimit)
{
    if (program_size > (1UL << 62))
        exitf("Program is too huge\n");
    active_thread.slot_occupation_status = SLOT_OCCUPIED;
}

/* No only will it launch a wave of deallocation in CA tree, but as a nice bonus it's
 * gonna deoccupy slot_occupation_status*/
void REGEX_IS024_Thread::delete_thread() noexcept {
    thread_print_debug(*this);
    my_assert(slot_occupation_status & SLOT_OCCUPIED);
    slot_occupation_status = SLOT_EMPTY_val;
    REGEX024_CollectionArrayNode* cur_CAptr = CAHptr;
    while (cur_CAptr){
        assert(cur_CAptr->refs > 0);
        if (--(cur_CAptr->refs) == 0){
            REGEX024_CollectionArrayNode* next_CAptr = cur_CAptr->prev;
            delete cur_CAptr;
            cur_CAptr = next_CAptr;
        } else
            break;
    }
    if (SAptr){
        if (--(SAptr[0]) == 0)
            free(SAptr);
    }
}

void emptify_one_of_new_read_halted_stacks(REGEX_IS024_CONTEXT& ctx, REGEX_IS024_Stack& type_new_stack){
    while (type_new_stack.non_empty()){
        REGEX_IS024_Thread& thread = ctx.READ_halted_slots[type_new_stack.pop()];
        assert(thread.slot_occupation_status & SLOT_OCCUPIED);
        thread.delete_thread();
    }
}

/* First it will try to pop pending thread from FORK_halted_stack
 * Then it will try popping thread from READ_halted_stack_old (checking if top
 * thread here is not actually SLOT_NEW). If something succeded, corresponding slot will be deoccupied, and
 * active slot will be occupied with it.
 *
 * try_to_continue_scheduled() assumes that active thread is unoccupied.*/
void REGEX_IS024_CONTEXT::try_to_continue_scheduled(){
    ctx_print_debug(*this);
    my_assert(!(active_thread.slot_occupation_status & SLOT_OCCUPIED));
    if (FORK_halted_stack.sz){
        regex_sslot_id_t ssid = FORK_halted_stack.pop();
        active_thread = FORK_halted_slots[ssid];
        FORK_halted_slots[ssid].slot_occupation_status = SLOT_EMPTY_val;
        return;
    }
    while (READ_halted_stack_old.sz){
        regex_sslot_id_t ssid = READ_halted_stack_old.pop();
        if (READ_halted_slots[ssid].slot_occupation_status & SLOT_NEW){
            /* This is the case when old thread was silently replaced by settled new thread */
            continue;
        }
        active_thread = READ_halted_slots[ssid];
        READ_halted_slots[ssid].slot_occupation_status = SLOT_EMPTY_val;
        return;
    }
    /* Failure here will be detected. We started with unoccupied active thread. iterator inside kick will see it */
}

void kick(REGEX_IS024_CONTEXT& ctx) {
    ctx_print_debug(ctx);
    while ((ctx.active_thread.slot_occupation_status & SLOT_OCCUPIED)
            && ctx.error == regex024_error_codes::stable){
        if (ctx.timer >= ctx.time_tick_limit)
            smitsya(timeout);
        ctx.timer++;

        check_available_prg(REGEX024_BYTECODE_INSTRUCTION_SZ)  // May return from kick(ctx)
        // smivanie from those instructions will be immediately detected. Everything is OK
        instruction_table(ctx);
    }
}


regex024_error_code REGEX_IS024_CONTEXT::feedSOF() {
    ctx_print_debug(*this);
    kick(*this);
    return error;
}

regex024_error_code REGEX_IS024_CONTEXT::startThread() {
    ctx_print_debug(*this);
    active_thread.slot_occupation_status = SLOT_OCCUPIED;
    active_thread.IP = unnatural_started_thread_IP;
    active_thread.SAptr = NULL;
    active_thread.CAHptr = NULL;
    kick(*this);
    return error;
}

/* I hate C++ (aka antichrist), won't use move sementic (aka drink cornsyrup) */
void swap_stacks(REGEX_IS024_Stack& A, REGEX_IS024_Stack& B) {
    std::swap(A.sz, B.sz);
    std::swap(A.slots, B.slots);
}

void fill_empty_old_read_halted_stack(REGEX_IS024_CONTEXT& ctx, REGEX_IS024_Stack& read_halted_stack_new){
    ctx_print_debug(ctx);
    my_assert(!ctx.READ_halted_stack_old.non_empty());

    // Actually, READ_halted_stack_old is always empty in this case
    assert(ctx.READ_halted_stack_old.empty());
    swap_stacks(ctx.READ_halted_stack_old, read_halted_stack_new);
    for (uint32_t i = 0; i < ctx.READ_halted_stack_old.sz; i++){
        REGEX_IS024_Thread& slot = ctx.READ_halted_slots[ctx.READ_halted_stack_old.slots[i]];
        /* Should get rid of 'NEW' qualifier */
        assert(slot.slot_occupation_status & SLOT_OCCUPIED);
        if (slot.slot_occupation_status & SLOT_OCCUPIED)
            slot.slot_occupation_status = SLOT_OCCUPIED;
    }
}

regex024_error_code REGEX_IS024_CONTEXT::feedCharacter(uint64_t input, uint64_t corresponding_byte_amount) {
    ctx_print_debug(*this);
    if (matched_thread.slot_occupation_status & SLOT_OCCUPIED)
        matched_thread.delete_thread();
    emptify_one_of_new_read_halted_stacks(*this, READ_halted_stack_new_second);
    fill_empty_old_read_halted_stack(*this, READ_halted_stack_new_first);
    INP = input;
    passed_bytes += corresponding_byte_amount;
    passed_chars++;
    try_to_continue_scheduled();
    kick(*this);
    return error;
}

regex024_error_code REGEX_IS024_CONTEXT::extendedFeedCharacter(uint64_t input) {
    ctx_print_debug(*this);
    if (matched_thread.slot_occupation_status & SLOT_OCCUPIED)
        matched_thread.delete_thread();
    fill_empty_old_read_halted_stack(*this, READ_halted_stack_new_second);
    INP = input;
    try_to_continue_scheduled();
    kick(*this);
    return error;
}

REGEX_IS024_CONTEXT::~REGEX_IS024_CONTEXT() {
    ctx_print_debug(*this);
    if (initialized){
        emptify_one_of_new_read_halted_stacks(*this, READ_halted_stack_new_first);
        emptify_one_of_new_read_halted_stacks(*this, READ_halted_stack_new_second);
        while (READ_halted_stack_old.non_empty()){
            REGEX_IS024_Thread& thread = READ_halted_slots[READ_halted_stack_old.pop()];
            assert(thread.slot_occupation_status & SLOT_OCCUPIED);
            if (!(thread.slot_occupation_status & SLOT_NEW))
                thread.delete_thread();
        }
        free(READ_halted_slots);
        while (FORK_halted_stack.non_empty())
            FORK_halted_slots[FORK_halted_stack.pop()].delete_thread();
        free(FORK_halted_slots);

        if (matched_thread.slot_occupation_status & SLOT_OCCUPIED){
            matched_thread.delete_thread();
        }
    }
}
